/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.env.Environment;
import org.ow2.bonita.env.EnvironmentFactory;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.def.element.impl.AttachmentDefinitionImpl;
import org.ow2.bonita.facade.exception.DocumentAlreadyExistsException;
import org.ow2.bonita.facade.exception.DocumentationCreationException;
import org.ow2.bonita.facade.runtime.impl.AttachmentInstanceImpl;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.identity.auth.DomainOwner;
import org.ow2.bonita.services.Document;
import org.ow2.bonita.services.DocumentationManager;
import org.ow2.bonita.services.LargeDataRepository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Anthony Birembaut, Matthieu Chaffotte
 */
public final class DbMigration {

  private static final Logger LOG = LoggerFactory.getLogger(DbMigration.class);

  private DbMigration() {
  }

  /**
   * Migrate DB schema of the database defined in the environment.
   */
  public static void migrateDb(final String domain, final String configurationName, final String db) throws Exception {
    BonitaConstants.getBonitaHomeFolder();
    final SessionFactory sessionFactory = DbTool.getSessionFactory(domain, configurationName.replaceAll("-configuration", "-session-factory"));
    final String database = db.toLowerCase();
    LOG.info("--------------- Running " + database + " " + configurationName + " DB migration ---------------");
    migrateAttachments(domain, sessionFactory);
    InputStream inputStream = null;
    try {
      inputStream = findMigrationScript(database);
      executeScript(sessionFactory, inputStream, database);
      LOG.info("---------------Schema migrated---------------");
      sessionFactory.close();
    } finally {
      if (inputStream != null) {
        inputStream.close();
      }
    }
  }

  public static void executeScript(SessionFactory sessionFactory, InputStream inputStream, String db) {
    final byte[] bytes = IoUtil.readBytes(inputStream);
    final String scriptContent = new String(bytes);
    final List<String> commands = getCommands(scriptContent, db);

    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();

    LOG.info("DB Commands Execution");
    for (String command : commands) {
      LOG.info("Executing command : " + command);
      try {
        session.createSQLQuery(command).executeUpdate();
      } catch (Exception e) {
        System.err.println("Error while executing command: " + command);
        LOG.error(e.getMessage(), e.getCause());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  public static List<String> getCommands(final String scriptContent, final String db) {
    String delimiter = ";";
    if (db.equals("sqlserver") || db.equals("sybase")) {
      delimiter = "go";
    }
    final String regex = delimiter.concat("\r?\n");
    final List<String> commands = new ArrayList<String>();
    final String[] tmp = scriptContent.split(regex);
    for (String command : tmp) {
      commands.add(command.trim());
    }
    int lastIndex = commands.size() - 1;
    String lastCommand = commands.get(lastIndex);
    int index = lastCommand.lastIndexOf(delimiter);
    if (index > 0) {
      lastCommand = lastCommand.substring(0, index);
      commands.remove(lastIndex);
      commands.add(lastCommand);
    }
    return commands;
  }

  public static InputStream findMigrationScript(final String db) {
    final StringBuilder migrationScript = new StringBuilder();
    migrationScript.append("/migration/").append(db).append(".sql");
    LOG.info("Loading Script " + migrationScript.toString());
    final InputStream inputStream = DbMigration.class.getResourceAsStream(migrationScript.toString());
    if (inputStream == null) {
      final String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_2");
      throw new IllegalArgumentException(message);
    }
    return inputStream;
  }

  public static void main(final String[] args) throws Exception {
    if (args == null || args.length != 3) {
      final String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_1");
      throw new IllegalArgumentException(message);
    }
    final String domain = args[0];
    final String db = args[1].toLowerCase();
    final boolean clean = Boolean.parseBoolean(args[2]);
    final String disableSearch = "bonita.search.use";
    System.setProperty(disableSearch, "false");
    if (clean) {
      cleanDocumentationManager(domain);
    }
    LOG.info("Starting History DB migration");
    migrateDb(domain, EnvConstants.HB_CONFIG_HISTORY, db);
    LOG.info("Starting Core DB migration");
    migrateDb(domain, EnvConstants.HB_CONFIG_CORE, db);
    System.clearProperty(disableSearch);
  }

  private static void cleanDocumentationManager(final String domain) throws Exception {
    final EnvironmentFactory envFactory = GlobalEnvironmentFactory.getEnvironmentFactory(domain);
    final Environment environment = envFactory.openEnvironment();
    try {
      final DocumentationManager manager = (DocumentationManager) envFactory.get(EnvConstants.DOCUMENTATION_MANAGER_DEFAULT_KEY);
      manager.clear();
    } finally {
      environment.close();
    }
  }

  @SuppressWarnings("unchecked")
  private static void migrateAttachments(final String domain, final SessionFactory sessionFactory) throws Exception {
    final EnvironmentFactory envFactory = GlobalEnvironmentFactory.getEnvironmentFactory(domain);
    final Environment environment = envFactory.openEnvironment();
    DomainOwner.setDomain(domain);
    final LargeDataRepository repository = (LargeDataRepository) envFactory.get(EnvConstants.LDR_SERVICE_DEFAULT_KEY);
    final DocumentationManager manager = (DocumentationManager) envFactory.get(EnvConstants.DOCUMENTATION_MANAGER_DEFAULT_KEY);
    final Session session = sessionFactory.openSession();
    try {
      final Transaction tx = session.beginTransaction();
      final Query defQuery = session.createSQLQuery("select PROC_UUID_ from BN_PROC_DEF");
      final List<String> definitionUUIDs = defQuery.list();
      if (definitionUUIDs != null) {
        for (String definitionUUID : definitionUUIDs) {
          final Query attDefQuery = session.createQuery("SELECT attachment FROM org.ow2.bonita.facade.def.element.impl.AttachmentDefinitionImpl AS attachment WHERE attachment.processDefinitionUUID = '" + definitionUUID + "'");
          final List<AttachmentDefinitionImpl> attachmentDefinitions = attDefQuery.list();
          migrateAttachmentDefinitions(new ProcessDefinitionUUID(definitionUUID), attachmentDefinitions, repository, manager);
        }
      }

      final Query instQuery = session.createSQLQuery("select INST_UUID_ from BN_PROC_INST");
      final List<String> instanceUUIDs = instQuery.list();
      if (instanceUUIDs != null) {
        for (String instanceUUID : instanceUUIDs) {
          final Query attQuery = session.createQuery("SELECT attachment FROM org.ow2.bonita.facade.runtime.impl.AttachmentInstanceImpl AS attachment WHERE processInstanceUUID = '" + instanceUUID + "' ORDER BY fileName, versionDate");
          final List<AttachmentInstanceImpl> attachments = attQuery.list();
          final ProcessInstanceUUID processInstanceUUID = new ProcessInstanceUUID(instanceUUID);
          migrateAttachmentInstances(processInstanceUUID.getProcessDefinitionUUID(), processInstanceUUID, attachments, repository, manager);
        }
      }

      tx.commit();
    } catch (HibernateException e) {
      LOG.warn("Unable to migrate attachments: ", e);
    } finally {
      DomainOwner.setDomain(null);
      environment.close();
      if (session != null) {
        session.close();
      }
    }
  }

  private static void migrateAttachmentInstances(ProcessDefinitionUUID definitionUUID, ProcessInstanceUUID instanceUUID, List<AttachmentInstanceImpl> attachments, LargeDataRepository repository, DocumentationManager manager)
  throws DocumentationCreationException {
    if (attachments != null) {
      String attachmentName = null;
      Document document = null;
      for (AttachmentInstanceImpl attachment : attachments) {
        final String name = attachment.getName();
        final String author = attachment.getAuthor();
        final Date versionDate = attachment.getVersionDate();
        final String key = Misc.getAttachmentIndexName(name, versionDate);
        final List<String> categories = Misc.getAttachmentCategories(instanceUUID);
        final byte[] content = repository.getData(byte[].class, categories, key);

        if (name.equals(attachmentName)) {
          if (content == null) {
            manager.createVersion(document.getId(), true, author, versionDate);
          } else {
            final String fileName = attachment.getFileName();
            String mimeType = getMimeType(attachment);
            manager.createVersion(document.getId(), true, author, versionDate, fileName, mimeType, content);
          }
        } else {
          attachmentName = attachment.getName();
          if (content == null) {
            document = manager.createDocument(name, definitionUUID, instanceUUID, author, versionDate);
          } else {
            final String fileName = attachment.getFileName();
            String mimeType = getMimeType(attachment);
            document = manager.createDocument(name, definitionUUID, instanceUUID, author, versionDate, fileName, mimeType, content);
          }
        }
      }
    }
  }

  private static String getMimeType(AttachmentInstanceImpl attachment) {
    final Map<String, String> metaData = attachment.getMetaData();                   
    String mimeType = metaData.get("content-type");
    if (mimeType == null) {
      mimeType = DocumentService.DEFAULT_MIME_TYPE;
    }
    return mimeType;
  }

  private static void migrateAttachmentDefinitions(final ProcessDefinitionUUID definitionUUID, List<AttachmentDefinitionImpl> attachments, LargeDataRepository repository, DocumentationManager manager)
  throws DocumentAlreadyExistsException, DocumentationCreationException {
    if (attachments!= null) {
      for (AttachmentDefinitionImpl attachment : attachments) {
        byte[] initialValue = null;
        final String name = attachment.getName();
        if (attachment.getFilePath() != null) {
          final List<String> categories = Misc.getAttachmentCategories(definitionUUID);
          initialValue = repository.getData(byte[].class, categories, attachment.getFilePath());
          repository.deleteData(categories, attachment.getFilePath());
          if (initialValue != null) {
            manager.createDocument(name, definitionUUID, null, attachment.getFileName(), DocumentService.DEFAULT_MIME_TYPE, initialValue);
          } else {
            manager.createDocument(name, definitionUUID, null);
          }
        } else {
          manager.createDocument(name, definitionUUID, null);
        }
      }
    }
  }

}
